#!/usr/bin/perl -w


#
#  Copyright (c) 2004-2005 The University of Chicago, as Operator
#  of Argonne National Laboratory.
#

# network_crawler.pl
#   Reads  accel_sw.log files,and writes resulting network data to db.
#
# Authors: Janet Anderson (jba@aps.anl.gov)
#
#
#

use Mail::Mailer;
use File::stat;
use Cwd;
use Blowfish_PP;

use LogUtil;
use NetworkCrawlerParser;
use NetworkCrawlerDBLayer;

my $status = 0;
my $allowed_updates = 4;

##########################################################
#  GET SCRIPT ARGUMENTS
##########################################################
my $test_mode = 0;
my $go = 0; # don't run script unless explicitly enabled
my $debug_level = "info"; # default
my $decryption_key = "";
my $replace_all_network_data = 0;
my $force_database_update = 0; # don't update database unless <= allowed_updates

if (!@ARGV) {
    print_usage();
    exit;
}
while (@ARGV) {
    $command = shift @ARGV;
    if ($command =~ /-help/) {
        print_usage();
        exit;
    } elsif  ($command =~ /--force/) {
        $force_database_update = 1;
    } elsif  ($command =~ /--all/) {
        $replace_all_network_data = 1;
    } elsif ($command =~ /--debug-level=(.*)/) {
        $debug_level = $1;
    } elsif ($command =~ /--go/) {
        $go = 1;
    } elsif ($command =~ /--key=(.*)/) {
        # Optional 16 byte blowfish decryption key. If given (non null),
        $decryption_key = $1;
    } elsif ($command =~ /--test/) {
        set_test_mode(1);
        $test_mode =1;
    } else {
        print_usage();
        exit;
    }
}
if (!$go ) {
    print "You must at a minimum use --go command line option to run the script!\n";
    exit;
}

##########
#  SETUP
##########

# Can be 'verbose', 'debug', 'info', 'warn', or 'error'
LogUtil::set_log_level($debug_level);

email_log('info',"Starting ioc network data update script.");   

# Set the input and temporary files
$infile            = "/usr/local/iocapps/iocinfo/network/accel_sw.log";
# =======  Debug settings
#$infile            = "./accel_sw.log";
# =======

$sb_in=stat($infile);
$infile_date=scalar localtime $sb_in->mtime;

################
#  Setup DBLayer
################

# Tell db layer where the home dir is before we do any chdir.
set_home_dir(getcwd());

# it is assumed the db.properties file is encrypted.
set_decryption_key($decryption_key);


#######################################################
#  Start database update or replace of ioc network data
#######################################################

my @network_hits=();

if ($replace_all_network_data) {

  #########################################################################
  # Replace all ioc network data with data from latest accel_sw.log file
  #########################################################################
  email_log('info',"New accel_sw.log file date: $infile_date");   
  #$status = zero_out_network_data(); # Some iocs have hand entered data which should not be zeroed out.
  if ( $status ==  0 ) {
    $status = get_all_network_data($infile,\@network_hits);
    if ( $status ==  0 ) {
      $status = load_new_network_data(\@network_hits);
    }
  }

} else {

  #############################################################################
  # Check accel_sw.log for differing network data
  #############################################################################

  email_log('info',"Updating modified network data.");   

  if ( $status == 0) {
    $networkData_ref = get_db_network_data();
    $status = get_modified_network_data($infile,\@network_hits,$networkData_ref);
    if ( $status >=  0 ) {
      if ( $status > $allowed_updates && !$force_database_update ) {
        email_log('error',"Update count $status is more than $allowed_updates. Database NOT updated.");   
      } else { 
        $status = load_new_network_data(\@network_hits);
      }
    }
  }
}

##################################
# Finish log, send email and exit.
##################################

email_log('info',"Finished ioc network data update script.");   
send_email();

exit;


#########################################################################
# Print out application usage text
#########################################################################
sub print_usage {
    print "\n";
    print "usage: network_crawler.pl --go [--key=<16 char key>] [--test/all/force] [--help] [--debug-level=[verbose|debug|info]]\n\n";
    print "The network_crawler reads a file of network switch interface data parsing\n";
    print "the data and comparing it to existing database data to determine if there\n";
    print "are changes to the ioc primary and secondary ethernet switch,blade, and \n";
    print "port settings. If there are less than $allowed_updates changes, the changes\n";
    print "are then inserted into the mysql database using perl DBI.\n";

    print "Log output goes to standard out, and can be adjusted from the default\n";
    print "\"info\" level. Watch out because verbose is VERY verbose.\n\n";

    print "You will need to establish the expected schema by running the\n";
    print "create_pv_tables.sql and alter_pv_tables.sql scripts in the rdbCore\n";
    print "ddl subdirectory. Then edit db.properties to set your local database\n";
    print "connection parameters (ie. user/password etc.).\n\n";

    print "The --test option runs the crawler, but does no db changes.\n";
    print "\n";
    print "The --all option replaces all ioc network data with data from the\n";
    print "latest accel_sw.log file.\n";
    print "\n";
    print "The --force option forces a database update when there are more\n";
    print "than the allowed number of database updates ($allowed_updates).\n";
    print "\n";

    print "You must have --go command line option to run the script!\n\n";
}

